//-----------------------------------------------------------------------------
// PS2IntDecoder.c
//
// This program decodes the key strokes from a PS/2 keyboard.
// It echoes each pressed key to the UART terminal and stores on a rx buffer.
// When the ENTER key is pressed, a line feed together with the content of the
// rx buffer are sent to the UART terminal, as a demonstration purpose.
//
// It doesnt sends commands to the keyboard; SHIFT, CAPS LOCK and ENTER keys
// are functional as expected when using a keyboard.
//-----------------------------------------------------------------------------
// This program is based on the following:
//
// Original title: keyboard_pc_to_rs232.c
// Description: Interfase entre un teclado convencional tipo AT y un puerto RS232C
// Date: Abr-2005
// Ver.Rev.: V01
// Author: XP8100 (xp8100@gmail.com)
//
// Based on the AN AVR313: Interfacing the PC AT Keyboard from ATMEL
// Adaptado para 16F628A por Redraven
//
// Link (in Spanish):
// http://picmania.garcia-cuervo.net/proyectos_teclado_ps2.htm
// Describes the functioning of the Ps/2 keyboard, codes, commands and connections.
// Also describes in detail the communication protocol and implements the code for
// a PIC 16f628.
//
//-----------------------------------------------------------------------------
// Adapted for the MikroC compiler and PIC18F45K22 by AlexBG 040320
//
// The array of n rows by 3 columns for the scan codes, comes from:
// http://www.lucadavidian.com/2017/11/15/interfacing-ps2-keyboard-to-a-microcontroller/
//
// The code for using two pointers to the buffer comes from:
// https://libstock.mikroe.com/projects/view/641/serial-to-glcd-for-easypic7
//-----------------------------------------------------------------------------
// Hardware and connections
// You can use the Mikroe PS/2 click module or build your own based on the schematic
// for this module; but for this program, do not connect it to the mikro BUS. Use the
// connections described below.
//
// https://www.mikroe.com/ps2-click
//
// PS/2 Module  EasyPIC 7
// --------------------------
// CS (CLK)       RB0
// RST (DATA)     RB3
// VCC            +5V
// GND            GND
//
// The reason for these connections is because the sample code provided for the
// Mikroe PS/2 click module, does not makes use of the external interrupt INT,
// located at RB0. Very probable, it is polling the clock signal assigned to
// pin CS of the mikro BUS standard.

// Instead of that, this program does use the external interrupt INT, located
// at RB0.
//
// On each falling edge of the clock signal, an interrupt is generated and at
// that moment, the level at RB3 pin is checked, for decoding the corresponding
// value (8 bits) for the pressed key.
//
// EasyPIC V7 settings:
// Enable the USB UART terminal with J3 and J4.
// Select the ON position for SW1.1 and SW2.1
// Connect your PS/2 click module to any of the external PORTB connectors.
//
//-----------------------------------------------------------------------------
//
// THIS DOCUMENT IS PROVIDED TO THE USER 'AS IS' WITHOUT ANY SPECIFIC GUARANTEE
// TO THE INFORMATION GIVEN ON IT.
//
//-----------------------------------------------------------------------------
#define TRUE 1
#define FALSE 0
//-----------------------------------------------------------------------------
// Global definitions
//-----------------------------------------------------------------------------
unsigned short dataReceived;          // For knowing when all the bits have been received
unsigned short keyData;               // Here, we store the byte corresponding to the pressed key
unsigned short rxBuffer[256];         // 256 byte rx buffer for storing the keys values
unsigned short inPoint;               // Pointer to the rx buffer
unsigned short outPoint;              // Pointer for flushing the rx buffer
unsigned short shifted = 0;           // For detecting when the SHIFT key has been pressed
unsigned short capslock = 0;          // For detecting when the CAPS LOCK has been pressed
unsigned short released = 0;          // For detecting when any key has been released
unsigned short anotherkey = 0;        // For detecting when any key is being pressed together with SHIFT

//-----------------------------------------------------------------------------
// Functions definitions
//-----------------------------------------------------------------------------
char Decode(unsigned char sc);        // Decode the received byte
void StoreKeyData(char ValueToStore); // Store the decoded data in a buffer
void TransmitrxBuffer(void);          // Send the received characters to the UART terminal

//-----------------------------------------------------------------------------
// Key definitions
//-----------------------------------------------------------------------------
const char scan_codes[][3] =
{
   0x0e,'|',' ',
   0x15,'q','Q',
   0x16,'1','!',
   0x1a,'z','Z',
   0x1b,'s','S',
   0x1c,'a','A',
   0x1d,'w','W',
   0x1e,'2','"',
   0x21,'c','C',
   0x22,'x','X',
   0x23,'d','D',
   0x24,'e','E',
   0x25,'4','$',
   0x26,'3','',
   0x29,' ',' ',
   0x2a,'v','V',
   0x2b,'f','F',
   0x2c,'t','T',
   0x2d,'r','R',
   0x2e,'5','%',
   0x31,'n','N',
   0x32,'b','B',
   0x33,'h','H',
   0x34,'g','G',
   0x35,'y','Y',
   0x36,'6','&',
   0x39,',',';',
   0x3a,'m','M',
   0x3b,'j','J',
   0x3c,'u','U',
   0x3d,'7','/',
   0x3e,'8',' ',
   0x41,',',';',
   0x42,'k','K',
   0x43,'i','I',
   0x44,'o','O',
   0x45,'0','=',
   0x46,'9',' ',
   0x49,'.',':',
   0x4a,'-','_',
   0x4b,'l','L',
   0x4c,'',' ',
   0x4d,'p','P',
   0x4e,'+','*',
   0x52,'',' ',
   0x54,'',' ',
   0x55,'\\',' ',
   0x5a,'\r','\r',
   0x5b,'',' ',
   0x5d,'\'',' ',
   0x61,'<','>',
   0x66,'\b','\b',
   0x69,'1',' ',
   0x6b,'4',' ',
   0x6c,'7',' ',
   0x70,'0',' ',
   0x71,',',' ',
   0x72,'2',' ',
   0x73,'5',' ',
   0x74,'6',' ',
   0x75,'8',' ',
   0x79,'+',' ',
   0x7a,'3',' ',
   0x7b,'-',' ',
   0x7c,'*',' ',
   0x7d,'9',' ',
     0,'\0','\0',
};

//-----------------------------------------------------------------------------
// ISR
// The clock signal from the keyboard is connected to RB0. An external interrupt
// is generated with each falling edge of the clock signal.
//-----------------------------------------------------------------------------
void interrupt(){
static unsigned short bitCount = 11;  // Used for counting the number of bits received

  if (INT0F_bit){
    //The bits 3 to 10 are data. Parity, Start and Stop are ignored
    if(bitCount < 11 && bitCount > 2){// If we are receiving data bits
      keyData = keyData >> 1;         // Shift to the right the decoded bits
      if(PORTB.B3){                   // Verify the level of the received data bit
        keyData = keyData | 0x80;     // if HIGH, save one on the MSB
      }
    }
    bitCount = bitCount - 1;          // Prepare for the next bit
    if(bitCount == 0){                // All the eight bits have been received
      bitCount = 11;                  // Reset the bit counting
      dataReceived = TRUE;            // Enable for decoding the received byte
    }
    INT0F_bit = 0;                    // Clear interrupt flag
  }
}

//-----------------------------------------------------------------------------
// Main
//-----------------------------------------------------------------------------
void main(void){
  char decodedKey;                    // Receives the decoded key

  ANSELB = 0;                         // Configure AN pins as digital
  ANSELC = 0;                         // Configure AN pins as digital
  ANSELD = 0;                         // Configure AN pins as digital
  SLRCON = 0;                         // Configure all PORTS at the standard Slew Rate
  TRISB = 0b00001001;                 // RB0, RB3 as inputs.

  // Enable the pullups located in the pins of PORTB
  WPUB = 0b00001001 ;                 // Enable individual pull-ups on input pins
  INTCON2.RBPU = 0;                   // Enable the pull-ups on pins on PORTB
                                      // configured as inputs and whose
                                      // corresponding bit in WPUB are set
  Delay_us(5);                        // Wait for the pullups
  // Set the external interrupt at RB0 for a falling edge
  INTEDG0_bit = 0;                    // Set interrupt on falling edge
  INT0IF_bit  = 0;                    // Clear INT0 flag
  INT0IE_bit  = 1;                    // Enable INT0 interrupts
  GIE_bit = 1;                        // Enable GLOBAL interrupts

  UART1_Init(9600);                   // Initialize UART module at 9600 bps
  Delay_ms(100);                      // Wait for UART module to stabilize

  dataReceived = FALSE;               // False meanwhile a key is decoded
  keyData=0;                          // Initialize for start decoding the pressed key
  inPoint=0;                          // Initialize the pointer to the buffer
  outPoint=0;                         // Initialize the flush pointer to the buffer

  UART1_Write_Text("\r\nPC AT Keyboard Interface Ver 1.0 by XP8100\r\n");
  UART1_Write_Text("Adapted for PIC by Redraven\r\n");
  UART1_Write_Text("Adapted for MikroC and PIC18F45K22 by AlexBG\r\n");
  UART1_Write_Text("PS/2 keyboard monitoring and decoder enabled...\r\n");

  while(1){
    if(dataReceived){                 // If the key data has been decoded at the INT
      switch(keyData){                // Check the key code
        // For the following three keys, decode but not print anything at the
        // UART terminal. Consideration for the capital letters, will be
        // done at the decode function.
        case 0x12:                    // Right shift
        case 0x59:                    // Left shift
          shifted = 1;
          break;
        case 0x58:                    // Caps lock
          capslock = ~ capslock;
          break;
        // Enter key. Print a carriage return and line feed.
        // For demonstration purposes, send the content of the rxbuffer to the UART.
        case 0x5A:                    // Enter key
          UART1_Write(13);            // ASCII carriage return
          UART1_Write(10);            // ASCII line feed (new line)
          TransmitrxBuffer();         // Send the content of the rxbuffer to the UART
          break;
        case 0xF0:                    // Break code (key has been released)
          if(shifted && anotherkey)   // Shift key is pressed and another key was
            anotherkey = 0;           // released
          else
            if(shifted)               // Key and shift released?
              shifted = 0;            // Yes, bring back to no shift key pressed
          break;
        // Rest of the keys. Decode and store on the rxbuffer. Print at the UART terminal.
        default:
          if(shifted)                 // Shift key pressed together with another one
            anotherkey = 1;
          decodedKey = Decode(keyData);
          StoreKeyData(decodedKey);
          UART1_Write(decodedKey);    // Echo to the UART
          break;
      }
      Delay_ms(10);                   // Used for debouncing the keyboard
      keyData = 0x00;                 // Clean for obtaining a new scan code
      dataReceived = FALSE;           // Disable until a new complete scan code is obtained
    }
  }
}

//-----------------------------------------------------------------------------
// Key decoder
//-----------------------------------------------------------------------------
char Decode(unsigned short scan_code){
  int i;                              // For counter
  const char CapitalLetters[26][2] =
  {0x1C, 'A',
   0x32, 'B',
   0x21, 'C',
   0x23, 'D',
   0x24, 'E',
   0x2B, 'F',
   0x34, 'G',
   0x33, 'H',
   0x43, 'I',
   0x3B, 'J',
   0x42, 'K',
   0x4B, 'L',
   0x3A, 'M',
   0x31, 'N',
   0x44, 'O',
   0x4D, 'P',
   0x15, 'Q',
   0x2D, 'R',
   0x1B, 'S',
   0x2C, 'T',
   0x3C, 'U',
   0x2A, 'V',
   0x1D, 'W',
   0x22, 'X',
   0x35, 'Y',
   0x1A, 'Z'
   };

  if(capslock){                       // If CAPS LOCK detected,
    for (i=0; i<26; i++){             // Look for a capital letter only
      if(scan_code == CapitalLetters[i][0])
        return(CapitalLetters[i][1]); // return capital letter
    }
  }
  for (i=0; i<67; i++){               // Look for the character
    if(scan_code == scan_codes[i][0])
      break;
  }
  if(shifted){                        // If SHIFT key detected,
    return(scan_codes[i][2]);         // return corresponding symbol or capital letter
  }
  else{                               // Non CAPS LOCK or SHIFT keys pressed,
    return(scan_codes[i][1]);         // return corresponding letter or symbol assigned
  }                                   // to the pressed key
}

//-----------------------------------------------------------------------------
// Store the decoded key value to the rx buffer
//-----------------------------------------------------------------------------
void StoreKeyData(char valueToStore){
  rxBuffer[inPoint]=valueToStore;
  inPoint++;                          // Increment the pointer at the rx buffer
}

//-----------------------------------------------------------------------------
// Send the contents of the rx buffer to the UART terminal
//-----------------------------------------------------------------------------
void TransmitrxBuffer(void){
  while(outPoint != inPoint){         // Do until both pointers are at the end of the data
    UART1_Write(rxBuffer[outPoint]);  // Send to the UART each character
    rxBuffer[outPoint]=" ";           // Delete the printed character
    outPoint++;                       // Increment printing pointer
  }
  UART1_Write(13);                    // ASCII carriage return
  UART1_Write(10);                    // ASCII line feed (new line)
  UART1_Write(13);                    // ASCII carriage return
  UART1_Write(10);                    // ASCII line feed (new line)
  inPoint = 0;                        // Initialize the storing pointer to the start of the buffer
  outPoint = 0;                       // Initialize the printing pointer
}